'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

const config = {
    // whether or not blockdom should normalize DOM whenever a block is created.
    // Normalizing dom mean removing empty text nodes (or containing only spaces)
    shouldNormalizeDom: true,
    // this is the main event handler. Every event handler registered with blockdom
    // will go through this function, giving it the data registered in the block
    // and the event
    mainEventHandler: (data, ev) => {
        if (typeof data === "function") {
            data(ev);
        }
        else if (Array.isArray(data)) {
            data[0](data[1], ev);
        }
    },
};

// -----------------------------------------------------------------------------
// Toggler node
// -----------------------------------------------------------------------------
class VToggler {
    constructor(key, child) {
        this.key = key;
        this.child = child;
    }
    mount(parent, afterNode) {
        this.parentEl = parent;
        this.child.mount(parent, afterNode);
    }
    moveBefore(other, afterNode) {
        this.child.moveBefore(other ? other.child : null, afterNode);
    }
    patch(other, withBeforeRemove) {
        if (this === other) {
            return;
        }
        let child1 = this.child;
        let child2 = other.child;
        if (this.key === other.key) {
            child1.patch(child2, withBeforeRemove);
        }
        else {
            child2.mount(this.parentEl, child1.firstNode());
            if (withBeforeRemove) {
                child1.beforeRemove();
            }
            child1.remove();
            this.child = child2;
            this.key = other.key;
        }
    }
    beforeRemove() { }
    remove() {
        this.child.remove();
    }
    firstNode() {
        return this.child.firstNode();
    }
    toString() {
        return this.child.toString();
    }
}
function toggler(key, child) {
    return new VToggler(key, child);
}

const { setAttribute, removeAttribute } = Element.prototype;
const tokenList = DOMTokenList.prototype;
const tokenListAdd = tokenList.add;
const tokenListRemove = tokenList.remove;
const isArray = Array.isArray;
const { split, trim } = String.prototype;
const wordRegexp = /\s+/;
/**
 * We regroup here all code related to updating attributes in a very loose sense:
 * attributes, properties and classs are all managed by the functions in this
 * file.
 */
function createAttrUpdater(attr) {
    return function (value) {
        if (value !== false) {
            setAttribute.call(this, attr, value === true ? "" : value);
        }
    };
}
function attrsSetter(attrs) {
    if (isArray(attrs)) {
        setAttribute.call(this, attrs[0], attrs[1]);
    }
    else {
        for (let k in attrs) {
            setAttribute.call(this, k, attrs[k]);
        }
    }
}
function attrsUpdater(attrs, oldAttrs) {
    if (isArray(attrs)) {
        const name = attrs[0];
        const val = attrs[1];
        if (name === oldAttrs[0]) {
            if (val === oldAttrs[1]) {
                return;
            }
            setAttribute.call(this, name, val);
        }
        else {
            removeAttribute.call(this, oldAttrs[0]);
            setAttribute.call(this, name, val);
        }
    }
    else {
        for (let k in oldAttrs) {
            if (!(k in attrs)) {
                removeAttribute.call(this, k);
            }
        }
        for (let k in attrs) {
            const val = attrs[k];
            if (val !== oldAttrs[k]) {
                setAttribute.call(this, k, val);
            }
        }
    }
}
function toClassObj(expr) {
    const result = {};
    switch (typeof expr) {
        case "string":
            // we transform here a list of classes into an object:
            //  'hey you' becomes {hey: true, you: true}
            const str = trim.call(expr);
            if (!str) {
                return {};
            }
            let words = split.call(str, wordRegexp);
            for (let i = 0, l = words.length; i < l; i++) {
                result[words[i]] = true;
            }
            return result;
        case "object":
            // this is already an object but we may need to split keys:
            // {'a': true, 'b c': true} should become {a: true, b: true, c: true}
            for (let key in expr) {
                const value = expr[key];
                if (value) {
                    const words = split.call(key, wordRegexp);
                    for (let word of words) {
                        result[word] = value;
                    }
                }
            }
            return result;
        case "undefined":
            return {};
        case "number":
            return { [expr]: true };
        default:
            return { [expr]: true };
    }
}
function setClass(val) {
    val = val === "" ? {} : toClassObj(val);
    // add classes
    const cl = this.classList;
    for (let c in val) {
        tokenListAdd.call(cl, c);
    }
}
function updateClass(val, oldVal) {
    oldVal = oldVal === "" ? {} : toClassObj(oldVal);
    val = val === "" ? {} : toClassObj(val);
    const cl = this.classList;
    // remove classes
    for (let c in oldVal) {
        if (!(c in val)) {
            tokenListRemove.call(cl, c);
        }
    }
    // add classes
    for (let c in val) {
        if (!(c in oldVal)) {
            tokenListAdd.call(cl, c);
        }
    }
}
function makePropSetter(name) {
    return function setProp(value) {
        this[name] = value;
    };
}
function isProp(tag, key) {
    switch (tag) {
        case "input":
            return (key === "checked" ||
                key === "indeterminate" ||
                key === "value" ||
                key === "readonly" ||
                key === "disabled");
        case "option":
            return key === "selected" || key === "disabled";
        case "textarea":
            return key === "readonly" || key === "disabled";
        case "button":
        case "select":
        case "optgroup":
            return key === "disabled";
    }
    return false;
}

function createEventHandler(event) {
    setupSyntheticEvent(event);
    const key = `__event__${event}`;
    return function setupHandler(data) {
        this[key] = data;
    };
}
function nativeToSyntheticEvent(event, name) {
    const eventKey = `__event__${name}`;
    let dom = event.target;
    while (dom !== null) {
        const data = dom[eventKey];
        if (data) {
            config.mainEventHandler(data, event);
            return;
        }
        dom = dom.parentNode;
    }
}
const CONFIGURED_SYNTHETIC_EVENTS = {};
function setupSyntheticEvent(name) {
    if (CONFIGURED_SYNTHETIC_EVENTS[name]) {
        return;
    }
    document.addEventListener(name, (event) => nativeToSyntheticEvent(event, name));
    CONFIGURED_SYNTHETIC_EVENTS[name] = true;
}

const getDescriptor$3 = (o, p) => Object.getOwnPropertyDescriptor(o, p);
const nodeProto$4 = Node.prototype;
const nodeInsertBefore$3 = nodeProto$4.insertBefore;
const nodeSetTextContent$1 = getDescriptor$3(nodeProto$4, "textContent").set;
const nodeRemoveChild$3 = nodeProto$4.removeChild;
// -----------------------------------------------------------------------------
// Multi NODE
// -----------------------------------------------------------------------------
class VMulti {
    constructor(children) {
        this.children = children;
    }
    mount(parent, afterNode) {
        const children = this.children;
        const l = children.length;
        const anchors = new Array(l);
        for (let i = 0; i < l; i++) {
            let child = children[i];
            if (child) {
                child.mount(parent, afterNode);
            }
            else {
                const childAnchor = document.createTextNode("");
                anchors[i] = childAnchor;
                nodeInsertBefore$3.call(parent, childAnchor, afterNode);
            }
        }
        this.anchors = anchors;
        this.parentEl = parent;
    }
    moveBefore(other, afterNode) {
        if (other) {
            const next = other.children[0];
            afterNode = (next ? next.firstNode() : other.anchors[0]) || null;
        }
        const children = this.children;
        const parent = this.parentEl;
        const anchors = this.anchors;
        for (let i = 0, l = children.length; i < l; i++) {
            let child = children[i];
            if (child) {
                child.moveBefore(null, afterNode);
            }
            else {
                const anchor = anchors[i];
                nodeInsertBefore$3.call(parent, anchor, afterNode);
            }
        }
    }
    patch(other, withBeforeRemove) {
        if (this === other) {
            return;
        }
        const children1 = this.children;
        const children2 = other.children;
        const anchors = this.anchors;
        const parentEl = this.parentEl;
        for (let i = 0, l = children1.length; i < l; i++) {
            const vn1 = children1[i];
            const vn2 = children2[i];
            if (vn1) {
                if (vn2) {
                    vn1.patch(vn2, withBeforeRemove);
                }
                else {
                    const afterNode = vn1.firstNode();
                    const anchor = document.createTextNode("");
                    anchors[i] = anchor;
                    nodeInsertBefore$3.call(parentEl, anchor, afterNode);
                    if (withBeforeRemove) {
                        vn1.beforeRemove();
                    }
                    vn1.remove();
                    children1[i] = undefined;
                }
            }
            else if (vn2) {
                children1[i] = vn2;
                const anchor = anchors[i];
                vn2.mount(parentEl, anchor);
                nodeRemoveChild$3.call(parentEl, anchor);
            }
        }
    }
    beforeRemove() {
        const children = this.children;
        for (let i = 0, l = children.length; i < l; i++) {
            const child = children[i];
            if (child) {
                child.beforeRemove();
            }
        }
    }
    remove() {
        const parentEl = this.parentEl;
        if (this.isOnlyChild) {
            nodeSetTextContent$1.call(parentEl, "");
        }
        else {
            const children = this.children;
            const anchors = this.anchors;
            for (let i = 0, l = children.length; i < l; i++) {
                const child = children[i];
                if (child) {
                    child.remove();
                }
                else {
                    nodeRemoveChild$3.call(parentEl, anchors[i]);
                }
            }
        }
    }
    firstNode() {
        const child = this.children[0];
        return child ? child.firstNode() : this.anchors[0];
    }
    toString() {
        return this.children.map((c) => c.toString()).join("");
    }
}
function multi(children) {
    return new VMulti(children);
}

const getDescriptor$2 = (o, p) => Object.getOwnPropertyDescriptor(o, p);
const nodeProto$3 = Node.prototype;
const characterDataProto$1 = CharacterData.prototype;
const nodeInsertBefore$2 = nodeProto$3.insertBefore;
const characterDataSetData$1 = getDescriptor$2(characterDataProto$1, "data").set;
const nodeRemoveChild$2 = nodeProto$3.removeChild;
class VText {
    constructor(text) {
        this.text = text;
    }
    mount(parent, afterNode) {
        this.parentEl = parent;
        const node = document.createTextNode(toText(this.text));
        nodeInsertBefore$2.call(parent, node, afterNode);
        this.el = node;
    }
    moveBefore(other, afterNode) {
        const target = other ? other.el : afterNode;
        nodeInsertBefore$2.call(this.parentEl, this.el, target);
    }
    patch(other) {
        const text2 = other.text;
        if (this.text !== text2) {
            characterDataSetData$1.call(this.el, toText(text2));
            this.text = text2;
        }
    }
    beforeRemove() { }
    remove() {
        nodeRemoveChild$2.call(this.parentEl, this.el);
    }
    firstNode() {
        return this.el;
    }
    toString() {
        return this.text;
    }
}
function text(str) {
    return new VText(str);
}
function toText(value) {
    switch (typeof value) {
        case "string":
            return value;
        case "number":
            return String(value);
        case "boolean":
            return value ? "true" : "false";
        default:
            return value || "";
    }
}

const getDescriptor$1 = (o, p) => Object.getOwnPropertyDescriptor(o, p);
const nodeProto$2 = Node.prototype;
const elementProto = Element.prototype;
const characterDataProto = CharacterData.prototype;
const characterDataSetData = getDescriptor$1(characterDataProto, "data").set;
const nodeGetFirstChild = getDescriptor$1(nodeProto$2, "firstChild").get;
const nodeGetNextSibling = getDescriptor$1(nodeProto$2, "nextSibling").get;
const NO_OP = () => { };
const cache = {};
/**
 * Compiling blocks is a multi-step process:
 *
 * 1. build an IntermediateTree from the HTML element. This intermediate tree
 *    is a binary tree structure that encode dynamic info sub nodes, and the
 *    path required to reach them
 * 2. process the tree to build a block context, which is an object that aggregate
 *    all dynamic info in a list, and also, all ref indexes.
 * 3. process the context to build appropriate builder/setter functions
 * 4. make a dynamic block class, which will efficiently collect references and
 *    create/update dynamic locations/children
 *
 * @param str
 * @returns a new block type, that can build concrete blocks
 */
function createBlock(str) {
    if (str in cache) {
        return cache[str];
    }
    // step 0: prepare html base element
    const doc = new DOMParser().parseFromString(str, "text/xml");
    const node = doc.firstChild;
    if (config.shouldNormalizeDom) {
        normalizeNode(node);
    }
    // step 1: prepare intermediate tree
    const tree = buildTree(node);
    // step 2: prepare block context
    const context = buildContext(tree);
    // step 3: build the final block class
    const template = tree.el;
    const Block = buildBlock(template, context);
    cache[str] = Block;
    return Block;
}
// -----------------------------------------------------------------------------
// Helper
// -----------------------------------------------------------------------------
function normalizeNode(node) {
    if (node.nodeType === Node.TEXT_NODE) {
        if (!/\S/.test(node.textContent)) {
            node.remove();
            return;
        }
    }
    if (node.nodeType === Node.ELEMENT_NODE) {
        if (node.tagName === "pre") {
            return;
        }
    }
    for (let i = node.childNodes.length - 1; i >= 0; --i) {
        normalizeNode(node.childNodes.item(i));
    }
}
function buildTree(node, parent = null, domParentTree = null) {
    switch (node.nodeType) {
        case 1: {
            // HTMLElement
            let isActive = false;
            const tagName = node.tagName;
            let el = undefined;
            const info = [];
            if (tagName.startsWith("block-text-")) {
                const index = parseInt(tagName.slice(11), 10);
                info.push({ type: "text", idx: index });
                el = document.createTextNode("");
                isActive = true;
            }
            if (tagName.startsWith("block-child-")) {
                domParentTree.forceRef = true;
                const index = parseInt(tagName.slice(12), 10);
                info.push({ type: "child", idx: index });
                el = document.createTextNode("");
                isActive = true;
            }
            if (!el) {
                el = document.createElement(tagName);
            }
            if (el instanceof HTMLElement) {
                const attrs = node.attributes;
                for (let i = 0; i < attrs.length; i++) {
                    const attrName = attrs[i].name;
                    const attrValue = attrs[i].value;
                    if (attrName.startsWith("block-handler-")) {
                        isActive = true;
                        const idx = parseInt(attrName.slice(14), 10);
                        info.push({
                            type: "handler",
                            idx,
                            event: attrValue,
                        });
                    }
                    else if (attrName.startsWith("block-attribute-")) {
                        isActive = true;
                        const idx = parseInt(attrName.slice(16), 10);
                        info.push({
                            type: "attribute",
                            idx,
                            name: attrValue,
                            tag: tagName,
                        });
                    }
                    else if (attrName === "block-attributes") {
                        isActive = true;
                        info.push({
                            type: "attributes",
                            idx: parseInt(attrValue, 10),
                        });
                    }
                    else if (attrName === "block-ref") {
                        isActive = true;
                        info.push({
                            type: "ref",
                            idx: parseInt(attrValue, 10),
                        });
                    }
                    else {
                        el.setAttribute(attrs[i].name, attrValue);
                    }
                }
            }
            const tree = {
                parent,
                firstChild: null,
                nextSibling: null,
                el,
                info,
                refN: isActive ? 1 : 0,
            };
            if (node.firstChild) {
                const childNode = node.childNodes[0];
                if (node.childNodes.length === 1 &&
                    childNode.nodeType === 1 &&
                    childNode.tagName.startsWith("block-child-")) {
                    const tagName = childNode.tagName;
                    const index = parseInt(tagName.slice(12), 10);
                    info.push({ idx: index, type: "child", isOnlyChild: true });
                    isActive = true;
                    tree.refN = 1;
                }
                else {
                    tree.firstChild = buildTree(node.firstChild, tree, tree);
                    el.appendChild(tree.firstChild.el);
                    let curNode = node.firstChild;
                    let curTree = tree.firstChild;
                    while ((curNode = curNode.nextSibling)) {
                        curTree.nextSibling = buildTree(curNode, curTree, tree);
                        el.appendChild(curTree.nextSibling.el);
                        curTree = curTree.nextSibling;
                    }
                }
            }
            if (isActive) {
                let cur = tree;
                while ((cur = cur.parent)) {
                    cur.refN++;
                }
            }
            return tree;
        }
        case 3:
        case 8: {
            // text node or comment node
            const el = node.nodeType === 3
                ? document.createTextNode(node.textContent)
                : document.createComment(node.textContent);
            return {
                parent: parent,
                firstChild: null,
                nextSibling: null,
                el,
                info: [],
                refN: 0,
            };
        }
    }
    throw new Error("boom");
}
function parentTree(tree) {
    let parent = tree.parent;
    while (parent && parent.nextSibling === tree) {
        tree = parent;
        parent = parent.parent;
    }
    return parent;
}
function buildContext(tree, ctx, fromIdx, toIdx) {
    if (!ctx) {
        const children = new Array(tree.info.filter((v) => v.type === "child").length);
        ctx = { collectors: [], locations: [], children, cbRefs: [], refN: tree.refN };
        fromIdx = 0;
        tree.refN - 1;
    }
    if (tree.refN) {
        const initialIdx = fromIdx;
        const isRef = tree.forceRef || tree.info.length > 0;
        const firstChild = tree.firstChild ? tree.firstChild.refN : 0;
        const nextSibling = tree.nextSibling ? tree.nextSibling.refN : 0;
        //node
        if (isRef) {
            for (let info of tree.info) {
                info.refIdx = initialIdx;
            }
            tree.refIdx = initialIdx;
            updateCtx(ctx, tree);
            fromIdx++;
        }
        // right
        if (nextSibling) {
            const idx = fromIdx + firstChild;
            ctx.collectors.push({ idx, prevIdx: initialIdx, getVal: nodeGetNextSibling });
            buildContext(tree.nextSibling, ctx, idx);
        }
        // left
        if (firstChild) {
            ctx.collectors.push({ idx: fromIdx, prevIdx: initialIdx, getVal: nodeGetFirstChild });
            buildContext(tree.firstChild, ctx, fromIdx);
        }
    }
    return ctx;
}
function updateCtx(ctx, tree) {
    for (let info of tree.info) {
        switch (info.type) {
            case "text":
                ctx.locations.push({
                    idx: info.idx,
                    refIdx: info.refIdx,
                    setData: setText,
                    updateData: setText,
                });
                break;
            case "child":
                if (info.isOnlyChild) {
                    // tree is the parentnode here
                    ctx.children[info.idx] = {
                        parentRefIdx: info.refIdx,
                        isOnlyChild: true,
                    };
                }
                else {
                    // tree is the anchor text node
                    ctx.children[info.idx] = {
                        parentRefIdx: parentTree(tree).refIdx,
                        afterRefIdx: info.refIdx,
                    };
                }
                break;
            case "attribute": {
                const refIdx = info.refIdx;
                let updater;
                let setter;
                if (isProp(info.tag, info.name)) {
                    const setProp = makePropSetter(info.name);
                    setter = setProp;
                    updater = setProp;
                }
                else if (info.name === "class") {
                    setter = setClass;
                    updater = updateClass;
                }
                else {
                    setter = createAttrUpdater(info.name);
                    updater = setter;
                }
                ctx.locations.push({
                    idx: info.idx,
                    refIdx,
                    setData: setter,
                    updateData: updater,
                });
                break;
            }
            case "attributes":
                ctx.locations.push({
                    idx: info.idx,
                    refIdx: info.refIdx,
                    setData: attrsSetter,
                    updateData: attrsUpdater,
                });
                break;
            case "handler": {
                const setupHandler = createEventHandler(info.event);
                ctx.locations.push({
                    idx: info.idx,
                    refIdx: info.refIdx,
                    setData: setupHandler,
                    updateData: setupHandler,
                });
                break;
            }
            case "ref":
                ctx.cbRefs.push(info.idx);
                ctx.locations.push({
                    idx: info.idx,
                    refIdx: info.refIdx,
                    setData: setRef,
                    updateData: NO_OP,
                });
        }
    }
}
// -----------------------------------------------------------------------------
// building the concrete block class
// -----------------------------------------------------------------------------
function buildBlock(template, ctx) {
    let B = createBlockClass(template, ctx);
    if (ctx.cbRefs.length) {
        const refs = ctx.cbRefs;
        B = class extends B {
            remove() {
                super.remove();
                for (let ref of refs) {
                    let fn = this.data[ref];
                    fn(null);
                }
            }
        };
    }
    if (ctx.children.length) {
        B = class extends B {
            constructor(data, children) {
                super(data);
                this.children = children;
            }
        };
        B.prototype.beforeRemove = VMulti.prototype.beforeRemove;
        return (data, children = []) => new B(data, children);
    }
    return (data) => new B(data);
}
function createBlockClass(template, ctx) {
    const { refN, collectors, children } = ctx;
    const colN = collectors.length;
    ctx.locations.sort((a, b) => a.idx - b.idx);
    const locations = ctx.locations.map((loc) => ({
        refIdx: loc.refIdx,
        setData: loc.setData,
        updateData: loc.updateData,
    }));
    const locN = locations.length;
    const childN = children.length;
    const childrenLocs = children;
    const isDynamic = refN > 0;
    // these values are defined here to make them faster to lookup in the class
    // block scope
    const nodeCloneNode = nodeProto$2.cloneNode;
    const nodeInsertBefore = nodeProto$2.insertBefore;
    const elementRemove = elementProto.remove;
    return class Block {
        constructor(data) {
            this.data = data;
        }
        beforeRemove() { }
        remove() {
            elementRemove.call(this.el);
        }
        firstNode() {
            return this.el;
        }
        moveBefore(other, afterNode) {
            const target = other ? other.el : afterNode;
            nodeInsertBefore.call(this.parentEl, this.el, target);
        }
        mount(parent, afterNode) {
            const el = nodeCloneNode.call(template, true);
            nodeInsertBefore.call(parent, el, afterNode);
            if (isDynamic) {
                // collecting references
                const refs = new Array(refN);
                this.refs = refs;
                refs[0] = el;
                for (let i = 0; i < colN; i++) {
                    const w = collectors[i];
                    refs[w.idx] = w.getVal.call(refs[w.prevIdx]);
                }
                // applying data to all update points
                if (locN) {
                    const data = this.data;
                    for (let i = 0; i < locN; i++) {
                        const loc = locations[i];
                        loc.setData.call(refs[loc.refIdx], data[i]);
                    }
                }
                // preparing all children
                if (childN) {
                    const children = this.children;
                    for (let i = 0; i < childN; i++) {
                        const child = children[i];
                        if (child) {
                            const loc = childrenLocs[i];
                            const afterNode = loc.afterRefIdx ? refs[loc.afterRefIdx] : null;
                            child.isOnlyChild = loc.isOnlyChild;
                            child.mount(refs[loc.parentRefIdx], afterNode);
                        }
                    }
                }
            }
            this.el = el;
            this.parentEl = parent;
        }
        patch(other, withBeforeRemove) {
            if (this === other) {
                return;
            }
            const refs = this.refs;
            // update texts/attributes/
            if (locN) {
                const data1 = this.data;
                const data2 = other.data;
                for (let i = 0; i < locN; i++) {
                    const val1 = data1[i];
                    const val2 = data2[i];
                    if (val1 !== val2) {
                        const loc = locations[i];
                        loc.updateData.call(refs[loc.refIdx], val2, val1);
                    }
                }
                this.data = data2;
            }
            // update children
            if (childN) {
                let children1 = this.children;
                const children2 = other.children;
                for (let i = 0; i < childN; i++) {
                    const child1 = children1[i];
                    const child2 = children2[i];
                    if (child1) {
                        if (child2) {
                            child1.patch(child2, withBeforeRemove);
                        }
                        else {
                            if (withBeforeRemove) {
                                child1.beforeRemove();
                            }
                            child1.remove();
                            children1[i] = undefined;
                        }
                    }
                    else if (child2) {
                        const loc = childrenLocs[i];
                        const afterNode = loc.afterRefIdx ? refs[loc.afterRefIdx] : null;
                        child2.mount(refs[loc.parentRefIdx], afterNode);
                        children1[i] = child2;
                    }
                }
            }
        }
        toString() {
            const div = document.createElement("div");
            this.mount(div, null);
            return div.innerHTML;
        }
    };
}
function setText(value) {
    characterDataSetData.call(this, toText(value));
}
function setRef(fn) {
    fn(this);
}

const getDescriptor = (o, p) => Object.getOwnPropertyDescriptor(o, p);
const nodeProto$1 = Node.prototype;
const nodeInsertBefore$1 = nodeProto$1.insertBefore;
const nodeAppendChild = nodeProto$1.appendChild;
const nodeRemoveChild$1 = nodeProto$1.removeChild;
const nodeSetTextContent = getDescriptor(nodeProto$1, "textContent").set;
// -----------------------------------------------------------------------------
// List Node
// -----------------------------------------------------------------------------
class VList {
    constructor(children) {
        this.children = children;
    }
    mount(parent, afterNode) {
        const children = this.children;
        const _anchor = document.createTextNode("");
        this.anchor = _anchor;
        nodeInsertBefore$1.call(parent, _anchor, afterNode);
        const l = children.length;
        if (l) {
            const mount = children[0].mount;
            for (let i = 0; i < l; i++) {
                mount.call(children[i], parent, _anchor);
            }
        }
        this.parentEl = parent;
    }
    moveBefore(other, afterNode) {
        if (other) {
            const next = other.children[0];
            afterNode = (next ? next.firstNode() : other.anchor) || null;
        }
        const children = this.children;
        for (let i = 0, l = children.length; i < l; i++) {
            children[i].moveBefore(null, afterNode);
        }
        this.parentEl.insertBefore(this.anchor, afterNode);
    }
    patch(other, withBeforeRemove) {
        if (this === other) {
            return;
        }
        const ch1 = this.children;
        const ch2 = other.children;
        if (ch2.length === 0 && ch1.length === 0) {
            return;
        }
        this.children = ch2;
        const proto = ch2[0] || ch1[0];
        const { mount: cMount, patch: cPatch, remove: cRemove, beforeRemove, moveBefore: cMoveBefore, firstNode: cFirstNode, } = proto;
        const _anchor = this.anchor;
        const isOnlyChild = this.isOnlyChild;
        const parent = this.parentEl;
        // fast path: no new child => only remove
        if (ch2.length === 0 && isOnlyChild) {
            if (withBeforeRemove) {
                for (let i = 0, l = ch1.length; i < l; i++) {
                    beforeRemove.call(ch1[i]);
                }
            }
            nodeSetTextContent.call(parent, "");
            nodeAppendChild.call(parent, _anchor);
            return;
        }
        let startIdx1 = 0;
        let startIdx2 = 0;
        let startVn1 = ch1[0];
        let startVn2 = ch2[0];
        let endIdx1 = ch1.length - 1;
        let endIdx2 = ch2.length - 1;
        let endVn1 = ch1[endIdx1];
        let endVn2 = ch2[endIdx2];
        let mapping = undefined;
        // let noFullRemove = this.hasNoComponent;
        while (startIdx1 <= endIdx1 && startIdx2 <= endIdx2) {
            // -------------------------------------------------------------------
            if (startVn1 === null) {
                startVn1 = ch1[++startIdx1];
                continue;
            }
            // -------------------------------------------------------------------
            if (endVn1 === null) {
                endVn1 = ch1[--endIdx1];
                continue;
            }
            // -------------------------------------------------------------------
            let startKey1 = startVn1.key;
            let startKey2 = startVn2.key;
            if (startKey1 === startKey2) {
                cPatch.call(startVn1, startVn2, withBeforeRemove);
                ch2[startIdx2] = startVn1;
                startVn1 = ch1[++startIdx1];
                startVn2 = ch2[++startIdx2];
                continue;
            }
            // -------------------------------------------------------------------
            let endKey1 = endVn1.key;
            let endKey2 = endVn2.key;
            if (endKey1 === endKey2) {
                cPatch.call(endVn1, endVn2, withBeforeRemove);
                ch2[endIdx2] = endVn1;
                endVn1 = ch1[--endIdx1];
                endVn2 = ch2[--endIdx2];
                continue;
            }
            // -------------------------------------------------------------------
            if (startKey1 === endKey2) {
                // bnode moved right
                cPatch.call(startVn1, endVn2, withBeforeRemove);
                ch2[endIdx2] = startVn1;
                const nextChild = ch2[endIdx2 + 1];
                cMoveBefore.call(startVn1, nextChild, _anchor);
                startVn1 = ch1[++startIdx1];
                endVn2 = ch2[--endIdx2];
                continue;
            }
            // -------------------------------------------------------------------
            if (endKey1 === startKey2) {
                // bnode moved left
                cPatch.call(endVn1, startVn2, withBeforeRemove);
                ch2[startIdx2] = endVn1;
                const nextChild = ch1[startIdx1];
                cMoveBefore.call(endVn1, nextChild, _anchor);
                endVn1 = ch1[--endIdx1];
                startVn2 = ch2[++startIdx2];
                continue;
            }
            // -------------------------------------------------------------------
            mapping = mapping || createMapping(ch1, startIdx1, endIdx1);
            let idxInOld = mapping[startKey2];
            if (idxInOld === undefined) {
                cMount.call(startVn2, parent, cFirstNode.call(startVn1) || null);
            }
            else {
                const elmToMove = ch1[idxInOld];
                cMoveBefore.call(elmToMove, startVn1, null);
                cPatch.call(elmToMove, startVn2, withBeforeRemove);
                ch2[startIdx2] = elmToMove;
                ch1[idxInOld] = null;
            }
            startVn2 = ch2[++startIdx2];
        }
        // ---------------------------------------------------------------------
        if (startIdx1 <= endIdx1 || startIdx2 <= endIdx2) {
            if (startIdx1 > endIdx1) {
                const nextChild = ch2[endIdx2 + 1];
                const anchor = nextChild ? cFirstNode.call(nextChild) || null : _anchor;
                for (let i = startIdx2; i <= endIdx2; i++) {
                    cMount.call(ch2[i], parent, anchor);
                }
            }
            else {
                for (let i = startIdx1; i <= endIdx1; i++) {
                    let ch = ch1[i];
                    if (ch) {
                        if (withBeforeRemove) {
                            beforeRemove.call(ch);
                        }
                        cRemove.call(ch);
                    }
                }
            }
        }
    }
    beforeRemove() {
        const children = this.children;
        const l = children.length;
        if (l) {
            const beforeRemove = children[0].beforeRemove;
            for (let i = 0; i < l; i++) {
                beforeRemove.call(children[i]);
            }
        }
    }
    remove() {
        const { parentEl, anchor } = this;
        if (this.isOnlyChild) {
            nodeSetTextContent.call(parentEl, "");
        }
        else {
            const children = this.children;
            const l = children.length;
            if (l) {
                const remove = children[0].remove;
                for (let i = 0; i < l; i++) {
                    remove.call(children[i]);
                }
            }
            nodeRemoveChild$1.call(parentEl, anchor);
        }
    }
    firstNode() {
        const child = this.children[0];
        return child ? child.firstNode() : undefined;
    }
    toString() {
        return this.children.map((c) => c.toString()).join("");
    }
}
function list(children) {
    return new VList(children);
}
function createMapping(ch1, startIdx1, endIdx2) {
    let mapping = {};
    for (let i = startIdx1; i <= endIdx2; i++) {
        mapping[ch1[i].key] = i;
    }
    return mapping;
}

const nodeProto = Node.prototype;
const nodeInsertBefore = nodeProto.insertBefore;
const nodeRemoveChild = nodeProto.removeChild;
class VHtml {
    constructor(html) {
        this.content = [];
        this.html = html;
    }
    mount(parent, afterNode) {
        this.parentEl = parent;
        const template = document.createElement("template");
        template.innerHTML = this.html;
        this.content = [...template.content.childNodes];
        for (let elem of this.content) {
            nodeInsertBefore.call(parent, elem, afterNode);
        }
        if (!this.content.length) {
            const textNode = document.createTextNode("");
            this.content.push(textNode);
            nodeInsertBefore.call(parent, textNode, afterNode);
        }
    }
    moveBefore(other, afterNode) {
        const target = other ? other.content[0] : afterNode;
        const parent = this.parentEl;
        for (let elem of this.content) {
            nodeInsertBefore.call(parent, elem, target);
        }
    }
    patch(other) {
        if (this === other) {
            return;
        }
        const html2 = other.html;
        if (this.html !== html2) {
            const parent = this.parentEl;
            // insert new html in front of current
            const afterNode = this.content[0];
            const template = document.createElement("template");
            template.innerHTML = html2;
            const content = [...template.content.childNodes];
            for (let elem of content) {
                nodeInsertBefore.call(parent, elem, afterNode);
            }
            if (!content.length) {
                const textNode = document.createTextNode("");
                content.push(textNode);
                nodeInsertBefore.call(parent, textNode, afterNode);
            }
            // remove current content
            this.remove();
            this.content = content;
        }
    }
    beforeRemove() { }
    remove() {
        const parent = this.parentEl;
        for (let elem of this.content) {
            nodeRemoveChild.call(parent, elem);
        }
    }
    firstNode() {
        return this.content[0];
    }
    toString() {
        return this.html;
    }
}
function html(str) {
    return new VHtml(str);
}

function mount(vnode, fixture) {
    vnode.mount(fixture, null);
}
function patch(vnode1, vnode2, withBeforeRemove = false) {
    vnode1.patch(vnode2, withBeforeRemove);
}
function remove(vnode, withBeforeRemove = false) {
    if (withBeforeRemove) {
        vnode.beforeRemove();
    }
    vnode.remove();
}
function withKey(vnode, key) {
    vnode.key = key;
    return vnode;
}

exports.config = config;
exports.createBlock = createBlock;
exports.html = html;
exports.list = list;
exports.mount = mount;
exports.multi = multi;
exports.patch = patch;
exports.remove = remove;
exports.text = text;
exports.toggler = toggler;
exports.withKey = withKey;
